/*
 * A database of m-sequence generators and support routines.
 *
 * Copyright 2005 Paul Millar
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 * Notes:
 *
 * M -sequences are maximum-length sequencies of numbers, so they
 * include every number (if we ignore 0). They are generated by taking
 * the previous number, left-shifting the bits, introducing a new bit
 * based on the existing ones. For example, for three bits we have
 * \x05 as the generator:
 *
 * 0x05 = 101  next
 *        ---  bit
 *        001  1    (= 1)
 *        011  1    (= 3)
 *        111  0    (= 7)
 *        110  1    (= 6)
 *        101  0    (= 5)
 *        010  0    (= 2)
 *        100  1    (= 4)
 */


#include <stdio.h>


unsigned long generator[] = {
  /* 0 (not useful, but needs initialising) */
   0, 

  /* 1 (again, not useful, but included) */
   0,

  /* 2 address lines */
   0x03,
  
  /* 3 address lines */
   0x05,

  /* 4 */
   0x09,

  /* 5 */
   0x12,

  /* 6 */
   0x21,

  /* 7 */
   0x41,

  /* 8 */
   0x8E,

  /* 9 */
   0x0108,

  /* 10 */
   0x0204,

  /* 11 */
   0x0402,

  /* 12 */
   0x0829,

  /* 13 */
   0x100D,

  /* 14 */
   0x2015,

  /* 15 */
   0x4001,

  /* 16 */
   0x8016,

  /* 17 */
   0x10004,

  /* 18 */
   0x20013,
};


/*
 * Updates the number to the next in M-seq order. Returns the value of the new
 * bit.
 */
int next_mseq_number( int order, unsigned long *number)
{
  int bit=0;
  unsigned long sig_bits;

  /* Treat zero as a special case: always clock-in a bit */
  if( *number == 0) {
    *number = 1;
    return 1;
  }

  for( sig_bits = *number & generator [order];
       sig_bits > 0;
       sig_bits >>=1) 
    if( sig_bits & 1)
      bit ^= 1;

  *number <<= 1;
  *number &= (1<<order)-1;
  *number |= bit;

  return bit;
} /* next_mseq_number */

